@@ -1,8 +1,8 @@
 # pylint: disable=unused-import
 from time import sleep
 import pytest
-import tests.helpers.utils as utils
 import tests.helpers.constants as constants
+from tests.helpers.utils import *
 from geckordp.rdp_client import RDPClient
 from geckordp.actors.root import RootActor
 from geckordp.actors.descriptors.tab import TabActor
@@ -113,8 +113,8 @@ def test_network_event():
         assert val != 0
 
         # get_response_cache
-        val = network_event.get_response_cache()["from"]
-        assert "netEvent" in val
+        val = network_event.get_response_cache()
+        assert response_valid("netEvent", val), str(val)
 
         # get_response_content
         val = network_event.get_response_content()["content"]["size"]
