@@ -21,3 +21,11 @@ def get_client_vars():
     descriptors = tab.get_target()
     browser = BrowsingContextActor(cl, descriptors["actor"])
     return cl, root, current_tab, tab, descriptors, browser
+
+
+def response_valid(actor_id: str, response: dict, allow_null=False) -> bool:
+    response_string = str(response).lower()
+    return (actor_id in response.get("from", "")
+            and "no such actor" not in response_string
+            and (allow_null or " is null" not in response_string)
+            and "unrecognized" not in response_string)
